""" Module containing a class that holds the tree search
"""
from __future__ import annotations

import json
from typing import TYPE_CHECKING

from aizynthfinder.chem.serialization import MoleculeDeserializer, MoleculeSerializer
from aizynthfinder.search.andor_trees import AndOrSearchTreeBase, SplitAndOrTree
from aizynthfinder.search.breadth_first.nodes import MoleculeNode
from aizynthfinder.utils.logging import logger

if TYPE_CHECKING:
    from aizynthfinder.chem import RetroReaction
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.reactiontree import ReactionTree
    from aizynthfinder.utils.type_utils import List, Optional, Sequence


class SearchTree(AndOrSearchTreeBase):
    """
    Encapsulation of the a breadth-first exhaustive search algorithm

    :ivar config: settings of the tree search algorithm
    :ivar root: the root node

    :param config: settings of the tree search algorithm
    :param root_smiles: the root will be set to a node representing this molecule, defaults to None
    """

    def __init__(
        self, config: Configuration, root_smiles: Optional[str] = None
    ) -> None:
        super().__init__(config, root_smiles)
        self._mol_nodes: List[MoleculeNode] = []
        self._added_mol_nodes: List[MoleculeNode] = []
        self._logger = logger()

        if root_smiles:
            self.root: Optional[MoleculeNode] = MoleculeNode.create_root(
                root_smiles, config
            )
            self._mol_nodes.append(self.root)
        else:
            self.root = None

        self._routes: List[ReactionTree] = []

        self.profiling = {
            "expansion_calls": 0,
            "reactants_generations": 0,
        }

    @classmethod
    def from_json(cls, filename: str, config: Configuration) -> SearchTree:
        """
        Create a new search tree by deserialization from a JSON file

        :param filename: the path to the JSON node
        :param config: the configuration of the search tree
        :return: a deserialized tree
        """

        def _find_mol_nodes(node):
            for child_ in node.children:
                tree._mol_nodes.append(child_)  # pylint: disable=protected-access
                for grandchild in child_.children:
                    _find_mol_nodes(grandchild)

        tree = cls(config)
        with open(filename, "r") as fileobj:
            dict_ = json.load(fileobj)
        mol_deser = MoleculeDeserializer(dict_["molecules"])
        tree.root = MoleculeNode.from_dict(dict_["tree"], config, mol_deser)
        tree._mol_nodes.append(tree.root)  # pylint: disable=protected-access
        for child in tree.root.children:
            _find_mol_nodes(child)
        return tree

    @property
    def mol_nodes(self) -> Sequence[MoleculeNode]:  # type: ignore
        """Return the molecule nodes of the tree"""
        return self._mol_nodes

    def one_iteration(self) -> bool:
        """
        Perform one iteration expansion.
        Expands all expandable molecule nodes in the tree, which should be
        on the same depth of the tree.

        :raises StopIteration: if the search should be pre-maturely terminated
        :return: if a solution was found
        :rtype: bool
        """
        if self.root is None:
            raise ValueError("Root is undefined. Cannot make an iteration")

        self._routes = []
        self._added_mol_nodes = []

        for next_node in self._mol_nodes:
            if next_node.expandable:
                self._expand(next_node)

        if not self._added_mol_nodes:
            self._logger.debug("No new nodes added in breadth-first iteration")
            raise StopIteration

        self._mol_nodes.extend(self._added_mol_nodes)
        solved = all(node.in_stock for node in self._mol_nodes if not node.children)
        return solved

    def routes(self) -> List[ReactionTree]:
        """
        Extracts and returns routes from the AND/OR tree

        :return: the routes
        """
        if self.root is None:
            return []
        if not self._routes:
            self._routes = SplitAndOrTree(self.root, self.config.stock).routes
        return self._routes

    def serialize(self, filename: str) -> None:
        """
        Seralize the search tree to a JSON file

        :param filename: the path to the JSON file
        :type filename: str
        """
        if self.root is None:
            raise ValueError("Cannot serialize tree as root is not defined")

        mol_ser = MoleculeSerializer()
        dict_ = {"tree": self.root.serialize(mol_ser), "molecules": mol_ser.store}
        with open(filename, "w") as fileobj:
            json.dump(dict_, fileobj, indent=2)

    def _expand(self, node: MoleculeNode) -> None:
        node.expandable = False
        reactions, _ = self.config.expansion_policy([node.mol])
        self.profiling["expansion_calls"] += 1

        if not reactions:
            return

        reactions_to_expand = []
        for reaction in reactions:
            try:
                self.profiling["reactants_generations"] += 1
                _ = reaction.reactants
            except:  # pylint: disable=bare-except
                continue
            if not reaction.reactants:
                continue
            for idx, _ in enumerate(reaction.reactants):
                rxn_copy = reaction.copy(idx)
                reactions_to_expand.append(rxn_copy)

        for rxn in reactions_to_expand:
            new_nodes = node.add_stub(rxn)
            self._added_mol_nodes.extend(new_nodes)



""" Module containing the implementation of the SC-score model for synthetic complexity scoring. """

from __future__ import annotations

import pickle
from typing import TYPE_CHECKING

import numpy as np
from rdkit.Chem import AllChem

from aizynthfinder.utils.math import (
    dense_layer_forward_pass,
    rectified_linear_unit,
    sigmoid,
)

if TYPE_CHECKING:
    from aizynthfinder.utils.type_utils import RdMol, Sequence, Tuple


class SCScore:
    """
    Encapsulation of the SCScore model

    Re-write of the SCScorer from the scscorer package

    The predictions of the score is made with a sanitized instance of an RDKit molecule

    .. code-block::

        mol = Molecule(smiles="CCC", sanitize=True)
        scscorer = SCScorer("path_to_model")
        score = scscorer(mol.rd_mol)

    The model provided when creating the scorer object should be pickled tuple.
    The first item of the tuple should be a list of the model weights for each layer.
    The second item of the tuple should be a list of the model biases for each layer.

    :param model_path: the filename of the model weights and biases
    :param fingerprint_length: the number of bits in the fingerprint
    :param fingerprint_radius: the radius of the fingerprint
    """

    def __init__(
        self,
        model_path: str,
        fingerprint_length: int = 1024,
        fingerprint_radius: int = 2,
    ) -> None:
        self._fingerprint_length = fingerprint_length
        self._fingerprint_radius = fingerprint_radius
        self._weights, self._biases = self._load_model(model_path)
        self.score_scale = 5.0

    def __call__(self, rd_mol: RdMol) -> float:
        fingerprint = self._make_fingerprint(rd_mol)
        normalized_score = self.forward(fingerprint)
        sc_score = (1 + (self.score_scale - 1) * normalized_score)[0]
        return sc_score

    # pylint: disable=invalid-name
    def forward(self, x: np.ndarray) -> np.ndarray:
        """Forward pass with dense neural network"""
        for weights, bias in zip(self._weights[:-1], self._biases[:-1]):
            x = dense_layer_forward_pass(x, weights, bias, rectified_linear_unit)
        return dense_layer_forward_pass(x, self._weights[-1], self._biases[-1], sigmoid)

    def _load_model(
        self, model_path: str
    ) -> Tuple[Sequence[np.ndarray], Sequence[np.ndarray]]:
        """Returns neural network model parameters."""
        with open(model_path, "rb") as fileobj:
            weights, biases = pickle.load(fileobj)

        weights = [np.asarray(item) for item in weights]
        biases = [np.asarray(item) for item in biases]
        return weights, biases

    def _make_fingerprint(self, rd_mol: RdMol) -> np.ndarray:
        """Returns the molecule's Morgan fingerprint"""
        fp_vec = AllChem.GetMorganFingerprintAsBitVect(
            rd_mol,
            self._fingerprint_radius,
            nBits=self._fingerprint_length,
            useChirality=True,
        )
        return np.array(
            fp_vec,
            dtype=bool,
        )



""" Module containing utility routines for MCTS. This is not part of public interface """
from __future__ import annotations

from typing import TYPE_CHECKING

from aizynthfinder.reactiontree import ReactionTreeLoader

if TYPE_CHECKING:
    from aizynthfinder.chem import RetroReaction
    from aizynthfinder.search.mcts import MctsNode
    from aizynthfinder.utils.type_utils import List, Optional, Tuple


class ReactionTreeFromSuperNode(ReactionTreeLoader):
    """
    Creates a reaction tree object from MCTS-like nodes and reaction objects
    """

    def _load(self, base_node: MctsNode) -> None:  # type: ignore
        actions, nodes = route_to_node(base_node)
        self.tree.created_at_iteration = base_node.created_at_iteration
        root_mol = nodes[0].state.mols[0]
        self._unique_mols[id(root_mol)] = root_mol.make_unique()
        self._add_node(
            self._unique_mols[id(root_mol)],
            in_stock=nodes[0].state.is_solved,
        )

        for child, action in zip(nodes[1:], actions):
            self._add_bipartite(child, action)

    def _add_bipartite(self, child: MctsNode, action: RetroReaction) -> None:
        reaction_obj = self._unique_reaction(action)
        self._add_node(reaction_obj, depth=2 * action.mol.transform + 1)
        self.tree.graph.add_edge(self._unique_mol(action.mol), reaction_obj)
        reactant_nodes = []
        for mol in child.state.mols:
            if mol.parent is action.mol:
                self._add_node(
                    self._unique_mol(mol),
                    depth=2 * mol.transform,
                    transform=mol.transform,
                    in_stock=mol in child.state.stock,
                )
                self.tree.graph.add_edge(reaction_obj, self._unique_mol(mol))
                reactant_nodes.append(self._unique_mol(mol))
        reaction_obj.reactants = (tuple(reactant_nodes),)


def route_to_node(
    from_node: MctsNode,
) -> Tuple[List[RetroReaction], List[MctsNode]]:
    """
    Return the route to a give node to the root.

    Will return both the actions taken to go between the nodes,
    and the nodes in the route themselves.

    :param from_node: the end of the route
    :return: the route
    """
    actions = []
    nodes = []
    current: Optional[MctsNode] = from_node

    while current is not None:
        parent = current.parent
        if parent is not None:
            action = parent[current]["action"]
            actions.append(action)
        nodes.append(current)
        current = parent
    return actions[::-1], nodes[::-1]



""" Module with script to download public data
"""
import argparse
import os
import sys

import requests
import tqdm

FILES_TO_DOWNLOAD = {
    "policy_model": {
        "filename": "uspto_model.hdf5",
        "url": "https://zenodo.org/record/7341155/files/uspto_keras_model.hdf5",
    },
    "policy_model_onnx": {
        "filename": "uspto_model.onnx",
        "url": "https://zenodo.org/record/7797465/files/uspto_model.onnx",
    },
    "template_file": {
        "filename": "uspto_templates.csv.gz",
        "url": "https://zenodo.org/record/7341155/files/uspto_unique_templates.csv.gz",
    },
    "ringbreaker_model": {
        "filename": "uspto_ringbreaker_model.hdf5",
        "url": "https://zenodo.org/record/7341155/files/uspto_ringbreaker_keras_model.hdf5",
    },
    "ringbreaker_model_onnx": {
        "filename": "uspto_ringbreaker_model.onnx",
        "url": "https://zenodo.org/record/7797465/files/uspto_ringbreaker_model.onnx",
    },
    "ringbreaker_templates": {
        "filename": "uspto_ringbreaker_templates.csv.gz",
        "url": "https://zenodo.org/record/7341155/files/uspto_ringbreaker_unique_templates.csv.gz",
    },
    "stock": {
        "filename": "zinc_stock.hdf5",
        "url": "https://ndownloader.figshare.com/files/23086469",
    },
    "filter_policy": {
        "filename": "uspto_filter_model.hdf5",
        "url": "https://ndownloader.figshare.com/files/25584743",
    },
    "filter_policy_onnx": {
        "filename": "uspto_filter_model.onnx",
        "url": "https://zenodo.org/record/7797465/files/uspto_filter_model.onnx",
    },
}

YAML_TEMPLATE = """expansion:
  uspto:
    - {}
    - {}
  ringbreaker:
    - {}
    - {}     
filter:
  uspto: {}
stock:
  zinc: {}
"""


def _download_file(url: str, filename: str) -> None:
    with requests.get(url, stream=True) as response:
        response.raise_for_status()
        total_size = int(response.headers.get("content-length", 0))
        pbar = tqdm.tqdm(
            total=total_size, desc=os.path.basename(filename), unit="B", unit_scale=True
        )
        with open(filename, "wb") as fileobj:
            for chunk in response.iter_content(chunk_size=1024):
                fileobj.write(chunk)
                pbar.update(len(chunk))
        pbar.close()


def main() -> None:
    """Entry-point for CLI"""
    parser = argparse.ArgumentParser("download_public_data")
    parser.add_argument(
        "path",
        default=".",
        help="the path to download the files",
    )
    path = parser.parse_args().path

    try:
        for filespec in FILES_TO_DOWNLOAD.values():
            _download_file(filespec["url"], os.path.join(path, filespec["filename"]))
    except requests.HTTPError as err:
        print(f"Download failed with message {str(err)}")
        sys.exit(1)

    with open(os.path.join(path, "config.yml"), "w") as fileobj:
        path = os.path.abspath(path)
        fileobj.write(
            YAML_TEMPLATE.format(
                os.path.join(path, FILES_TO_DOWNLOAD["policy_model_onnx"]["filename"]),
                os.path.join(path, FILES_TO_DOWNLOAD["template_file"]["filename"]),
                os.path.join(
                    path, FILES_TO_DOWNLOAD["ringbreaker_model_onnx"]["filename"]
                ),
                os.path.join(
                    path, FILES_TO_DOWNLOAD["ringbreaker_templates"]["filename"]
                ),
                os.path.join(path, FILES_TO_DOWNLOAD["filter_policy_onnx"]["filename"]),
                os.path.join(path, FILES_TO_DOWNLOAD["stock"]["filename"]),
            )
        )
    print("Configuration file written to config.yml")


if __name__ == "__main__":
    main()



""" Module containing a classes representation various tree nodes
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from aizynthfinder.chem import TreeMolecule
from aizynthfinder.chem.serialization import deserialize_action, serialize_action
from aizynthfinder.search.andor_trees import TreeNodeMixin
from aizynthfinder.search.retrostar.cost import MoleculeCost

if TYPE_CHECKING:
    from aizynthfinder.chem import RetroReaction
    from aizynthfinder.chem.serialization import (
        MoleculeDeserializer,
        MoleculeSerializer,
    )
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.utils.type_utils import List, Optional, Sequence, Set, StrDict


class MoleculeNode(TreeNodeMixin):
    """
    An OR node representing a molecule

    :ivar cost: the cost of synthesizing the molecule
    :ivar expandable: if True, this node is part of the frontier
    :ivar mol: the molecule represented by the node
    :ivar in_stock: if True the molecule is in stock and hence should not be expanded
    :ivar parent: the parent of the node
    :ivar solved: if True the molecule is in stock or at least one child node is solved
    :ivar value: the current rn(m|T)

    :param mol: the molecule to be represented by the node
    :param config: the configuration of the search
    :param parent: the parent of the node, optional
    """

    def __init__(
        self,
        mol: TreeMolecule,
        config: Configuration,
        molecule_cost: MoleculeCost,
        parent: Optional[ReactionNode] = None,
    ) -> None:
        self.mol = mol
        self._config = config
        self.molecule_cost = molecule_cost
        self.cost = self.molecule_cost(mol)
        self.value = self.cost
        self.in_stock = mol in config.stock
        self.parent = parent

        self._children: List[ReactionNode] = []
        self.solved = self.in_stock
        # Makes it unexpandable if we have reached maximum depth
        self.expandable = self.mol.transform < self._config.search.max_transforms

        if self.in_stock:
            self.expandable = False
            self.value = 0

    @classmethod
    def create_root(
        cls, smiles: str, config: Configuration, molecule_cost: MoleculeCost
    ) -> "MoleculeNode":
        """
        Create a root node for a tree using a SMILES.

        :param smiles: the SMILES representation of the root state
        :param config: settings of the tree search algorithm
        :return: the created node
        """
        mol = TreeMolecule(parent=None, transform=0, smiles=smiles)
        return MoleculeNode(mol=mol, config=config, molecule_cost=molecule_cost)

    @classmethod
    def from_dict(
        cls,
        dict_: StrDict,
        config: Configuration,
        molecules: MoleculeDeserializer,
        molecule_cost: MoleculeCost,
        parent: Optional[ReactionNode] = None,
    ) -> "MoleculeNode":
        """
        Create a new node from a dictionary, i.e. deserialization

        :param dict_: the serialized node
        :param config: settings of the tree search algorithm
        :param molecules: the deserialized molecules
        :param parent: the parent node
        :return: a deserialized node
        """
        mol = molecules.get_tree_molecules([dict_["mol"]])[0]
        node = MoleculeNode(mol, config, molecule_cost, parent)
        node.molecule_cost = molecule_cost
        for attr in ["cost", "expandable", "value"]:
            setattr(node, attr, dict_[attr])
        node.children = [
            ReactionNode.from_dict(
                child, config, molecules, node.molecule_cost, parent=node
            )
            for child in dict_["children"]
        ]
        return node

    @property  # type: ignore
    def children(self) -> List[ReactionNode]:  # type: ignore
        """Gives the reaction children nodes"""
        return self._children

    @children.setter
    def children(self, value: List[ReactionNode]) -> None:
        self._children = value

    @property
    def target_value(self) -> float:
        """
        The V_t(m|T) value,
        the current cost of the tree containing this node

        :return: the value
        """
        if self.parent:
            return self.parent.target_value
        return self.value

    @property
    def prop(self) -> StrDict:
        return {"solved": self.solved, "mol": self.mol}

    def add_stub(self, cost: float, reaction: RetroReaction) -> Sequence[MoleculeNode]:
        """
        Add a stub / sub-tree to this node

        :param cost: the cost of the reaction
        :param reaction: the reaction creating the stub
        :return: list of all newly added molecular nodes
        """
        reactants = reaction.reactants[reaction.index]
        if not reactants:
            return []

        ancestors = self.ancestors()
        for mol in reactants:
            if mol in ancestors:
                return []

        rxn_node = ReactionNode.create_stub(
            cost=cost,
            reaction=reaction,
            parent=self,
            config=self._config,
        )
        self._children.append(rxn_node)

        return rxn_node.children

    def ancestors(self) -> Set[TreeMolecule]:
        """
        Return the ancestors of this node

        :return: the ancestors
        :rtype: set
        """
        if not self.parent:
            return {self.mol}

        ancestors = self.parent.parent.ancestors()
        ancestors.add(self.mol)
        return ancestors

    def close(self) -> float:
        """
        Updates the values of this node after expanding it.

        :return: the delta V value
        :rtype: float
        """
        self.solved = any(child.solved for child in self.children)
        if self.children:
            new_value = np.min([child.value for child in self.children])
        else:
            new_value = np.inf

        v_delta = new_value - self.value
        self.value = new_value

        self.expandable = False
        return v_delta

    def serialize(self, molecule_store: MoleculeSerializer) -> StrDict:
        """
        Serialize the node object to a dictionary

        :param molecule_store: the serialized molecules
        :return: the serialized node
        """
        dict_ = {attr: getattr(self, attr) for attr in ["cost", "expandable", "value"]}
        dict_["mol"] = molecule_store[self.mol]
        dict_["children"] = [child.serialize(molecule_store) for child in self.children]
        return dict_

    def update(self, solved: bool) -> None:
        """
        Update the node as part of the update algorithm,
        calling the `update()` method of its parent if available.

        :param solved: if the child node was solved
        """
        new_value = np.min([child.value for child in self.children])
        new_solv = self.solved or solved
        updated = (self.value != new_value) or (self.solved != new_solv)

        v_delta = new_value - self.value
        self.value = new_value
        self.solved = new_solv

        if updated and self.parent:
            self.parent.update(v_delta, from_mol=self.mol)


class ReactionNode(TreeNodeMixin):
    """
    An AND node representing a reaction

    :ivar cost: the cost of the reaction
    :ivar parent: the parent of the node
    :ivar reaction: the reaction represented by the node
    :ivar solved: if True all children nodes are solved
    :ivar target_value: the V(m|T) for the children, the current cost
    :ivar value: the current rn(r|T)

    :param cost: the cost of the reaction
    :param reaction: the reaction to be represented by the node
    :param parent: the parent of the node
    """

    def __init__(
        self, cost: float, reaction: RetroReaction, parent: MoleculeNode
    ) -> None:
        self.parent = parent
        self.cost = cost
        self.reaction = reaction

        self._children: List[MoleculeNode] = []
        self.solved = False
        # rn(R|T)
        self.value = self.cost
        # V(R|T) = V(m|T) for m in children
        self.target_value = self.parent.target_value - self.parent.value + self.value

    @classmethod
    def create_stub(
        cls,
        cost: float,
        reaction: RetroReaction,
        parent: MoleculeNode,
        config: Configuration,
    ) -> ReactionNode:
        """
        Create a ReactionNode and creates all the MoleculeNode objects
        that are the children of the node.

        :param cost: the cost of the reaction
        :param reaction: the reaction to be represented by the node
        :param parent: the parent of the node
        :param config: the configuration of the search tree
        """
        node = cls(cost, reaction, parent)
        reactants = reaction.reactants[reaction.index]
        node.children = [
            MoleculeNode(
                mol=mol, config=config, molecule_cost=parent.molecule_cost, parent=node
            )
            for mol in reactants
        ]
        node.solved = all(child.solved for child in node.children)
        # rn(R|T)
        node.value = node.cost + sum(child.value for child in node.children)
        # V(R|T) = V(m|T) for m in children
        node.target_value = node.parent.target_value - node.parent.value + node.value
        return node

    @classmethod
    def from_dict(
        cls,
        dict_: StrDict,
        config: Configuration,
        molecules: MoleculeDeserializer,
        molecule_cost: MoleculeCost,
        parent: MoleculeNode,
    ) -> ReactionNode:
        """
        Create a new node from a dictionary, i.e. deserialization

        :param dict_: the serialized node
        :param config: the configuration of the tree search
        :param molecules: the deserialized molecules
        :param parent: the parent node
        :return: a deserialized node
        """
        reaction = deserialize_action(dict_["reaction"], molecules)
        node = cls(0, reaction, parent)
        for attr in ["cost", "value", "target_value"]:
            setattr(node, attr, dict_[attr])
        node.children = [
            MoleculeNode.from_dict(child, config, molecules, molecule_cost, parent=node)
            for child in dict_["children"]
        ]
        node.solved = all(child.solved for child in node.children)
        return node

    @property  # type: ignore
    def children(self) -> List[MoleculeNode]:  # type: ignore
        """Gives the molecule children nodes"""
        return self._children

    @children.setter
    def children(self, value: List[MoleculeNode]) -> None:
        self._children = value

    @property
    def prop(self) -> StrDict:
        return {"solved": self.solved, "reaction": self.reaction}

    def serialize(self, molecule_store: MoleculeSerializer) -> StrDict:
        """
        Serialize the node object to a dictionary

        :param molecule_store: the serialized molecules
        :return: the serialized node
        """
        dict_ = {
            attr: getattr(self, attr) for attr in ["cost", "value", "target_value"]
        }
        dict_["reaction"] = serialize_action(self.reaction, molecule_store)
        dict_["children"] = [child.serialize(molecule_store) for child in self.children]
        return dict_

    def update(self, value: float, from_mol: Optional[TreeMolecule] = None) -> None:
        """
        Update the node as part of the update algorithm,
        calling the `update()` method of its parent

        :param value: the delta V value
        :param from_mol: the molecule being expanded, used for excluding propagation
        """
        self.value += value
        self.target_value += value
        self.solved = all(node.solved for node in self.children)

        if value != 0:
            self._propagate(value, exclude=from_mol)

        self.parent.update(self.solved)

    def _propagate(self, value: float, exclude: Optional[TreeMolecule] = None) -> None:
        if not exclude:
            self.target_value += value

        for child in self.children:
            if exclude is None or child.mol is not exclude:
                for grandchild in child.children:
                    grandchild._propagate(value)  # pylint: disable=protected-access



""" Module contain a class that encapsulate the state of search tree node.
"""
from __future__ import annotations

import os
from typing import TYPE_CHECKING

from rdkit.Chem import Draw

from aizynthfinder.chem import TreeMolecule

if TYPE_CHECKING:
    from aizynthfinder.chem import MoleculeDeserializer, MoleculeSerializer
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.utils.type_utils import Any, List, Optional, Sequence, StrDict

os.environ["PYTHONHASHSEED"] = "42"


class MctsState:
    """
    Encapsulation of the molecular state of a node.

    A state consists of an immutable list of molecules that are either solved
    (can be found in stock) or that potentially can be expanded to new molecules
    by applying a reaction on them.

    The class is hashable and comparable by the inchi keys of all the molecules.

    :ivar mols: the list of molecules
    :ivar expandable_mols: the list of molecules not in stock
    :ivar stock: the configured stock
    :ivar in_stock_list: for each molecule if they are in stock
    :ivar is_solved: is true if all molecules are in stock:
    :ivar max_transforms: the maximum of the transforms of the molecule
    :ivar is_terminal: is true if the all molecules are in stock or if the maximum transforms has been reached
    :ivar expandables_hash: an hash string computed on the expandable molecules

    :param mols: the molecules of the state
    :param config: settings of the tree search algorithm
    """

    def __init__(self, mols: Sequence[TreeMolecule], config: Configuration) -> None:
        self.mols = mols
        self.stock = config.stock
        self.in_stock_list = [mol in self.stock for mol in self.mols]
        self.expandable_mols = [
            mol for mol, in_stock in zip(self.mols, self.in_stock_list) if not in_stock
        ]
        self._stock_availability: Optional[List[str]] = None
        self.is_solved = all(self.in_stock_list)
        self.max_transforms = max(mol.transform for mol in self.mols)
        self.is_terminal = (
            self.max_transforms >= config.search.max_transforms
        ) or self.is_solved

        inchis = [mol.inchi_key for mol in self.mols]
        self._hash = hash(tuple(sorted(inchis)))

        inchis = [mol.inchi_key for mol in self.expandable_mols]
        self.expandables_hash = hash(tuple(sorted(inchis)))

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MctsState):
            return False
        return self.__hash__() == other.__hash__()

    def __str__(self) -> str:
        """A string representation of the state (for print(state))"""
        string = (
            f"{str([mol.smiles for mol in self.mols])}\n"
            + f"{str([mol.transform for mol in self.mols])}\n"
            + f"{str([mol.parent.smiles if mol.parent else '-' for mol in self.mols])}\n"
            + f"{str(self.in_stock_list)}\n Solved: {self.is_solved}"
        )
        return string

    @classmethod
    def from_dict(
        cls, dict_: StrDict, config: Configuration, molecules: MoleculeDeserializer
    ) -> "MctsState":
        """
        Create a new state from a dictionary, i.e. deserialization

        :param dict_: the serialized state
        :type dict_: dict
        :param config: settings of the tree search algorithm
        :type config: Configuration
        :param molecules: the deserialized molecules
        :type molecules: MoleculeDeserializer
        :return: a deserialized state
        :rtype: State
        """
        mols = molecules.get_tree_molecules(dict_["mols"])
        return MctsState(mols, config)

    @property
    def stock_availability(self) -> List[str]:
        """
        Returns a list of availabilities for all molecules

        :return: the list
        :rtype: list of str
        """
        if not self._stock_availability:
            self._stock_availability = [
                self.stock.availability_string(mol) for mol in self.mols
            ]
        return self._stock_availability

    def serialize(self, molecule_store: MoleculeSerializer) -> StrDict:
        """
        Serialize the state object to a dictionary

        :param molecule_store: the serialized molecules
        :type molecule_store: MolecularSerializer
        :return: the serialized state
        :rtype: dict
        """
        return {"mols": [molecule_store[mol] for mol in self.mols]}

    def to_image(self, ncolumns: int = 6) -> Any:
        """
        Constructs an image representation of the state

        :param ncolumns: number of molecules per row, defaults to 6
        :type ncolumns: int, optional
        :return: the image representation
        :rtype: a PIL image
        """
        for mol in self.mols:
            mol.sanitize()
        legends = self.stock_availability
        mols = [mol.rd_mol for mol in self.mols]
        return Draw.MolsToGridImage(mols, molsPerRow=ncolumns, legends=legends)



""" Module containing a CLI for concatenating output files (hdf5/json.gz files)
"""
import argparse

from aizynthfinder.utils.files import cat_datafiles


def main() -> None:
    """Entry-point for the cat_aizynth_output CLI"""
    parser = argparse.ArgumentParser("cat_aizynthcli_output")
    parser.add_argument(
        "--files",
        required=True,
        nargs="+",
        help="the output filenames",
    )
    parser.add_argument(
        "--output",
        required=True,
        default="output.json.gz",
        help="the name of the concatenate output file ",
    )
    parser.add_argument(
        "--trees",
        help="if given, save all trees to this file",
    )
    args = parser.parse_args()

    cat_datafiles(args.files, args.output, args.trees)


if __name__ == "__main__":
    main()



""" Module containing helper classes and routines for serialization.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import aizynthfinder.chem
from aizynthfinder.utils.loading import load_dynamic_class

if TYPE_CHECKING:
    from aizynthfinder.chem import RetroReaction
    from aizynthfinder.utils.type_utils import Any, Dict, Optional, Sequence, StrDict


class MoleculeSerializer:
    """
    Utility class for serializing molecules

    The id of the molecule to be serialized can be obtained with:

    .. code-block::

        serializer = MoleculeSerializer()
        mol = Molecule(smiles="CCCO")
        idx = serializer[mol]

    which will take care of the serialization of the molecule.
    """

    def __init__(self) -> None:
        self._store: Dict[int, Any] = {}

    def __getitem__(self, mol: Optional[aizynthfinder.chem.Molecule]) -> Optional[int]:
        if mol is None:
            return None

        id_ = id(mol)
        if id_ not in self._store:
            self._add_mol(mol)
        return id_

    @property
    def store(self) -> Dict[int, Any]:
        """Return all serialized molecules as a dictionary"""
        return self._store

    def _add_mol(self, mol: aizynthfinder.chem.Molecule) -> None:
        id_ = id(mol)
        dict_ = {"smiles": mol.smiles, "class": mol.__class__.__name__}
        if isinstance(mol, aizynthfinder.chem.TreeMolecule):
            dict_["parent"] = self[mol.parent]
            dict_["transform"] = mol.transform
            if not mol.parent:
                dict_["smiles"] = mol.original_smiles
            else:
                dict_["smiles"] = mol.mapped_smiles
        self._store[id_] = dict_


class MoleculeDeserializer:
    """
    Utility class for deserializing molecules.
    The serialized molecules are created upon instantiation of the class.

    The deserialized molecules can be obtained with:

    .. code-block::

        deserializer = MoleculeDeserializer()
        mol = deserializer[idx]

    """

    def __init__(self, store: Dict[int, Any]) -> None:
        self._objects: Dict[int, Any] = {}
        self._create_molecules(store)

    def __getitem__(self, id_: Optional[int]) -> Optional[aizynthfinder.chem.Molecule]:
        if id_ is None:
            return None
        return self._objects[id_]

    def get_tree_molecules(
        self, ids: Sequence[int]
    ) -> Sequence[aizynthfinder.chem.TreeMolecule]:
        """
        Return multiple deserialized tree molecules

        :param ids: the list of IDs to deserialize
        :return: the molecule objects
        """
        objects = []
        for id_ in ids:
            obj = self[id_]
            if obj is None or not isinstance(obj, aizynthfinder.chem.TreeMolecule):
                raise ValueError(f"Failed to deserialize molecule with id {id_}")
            objects.append(obj)
        return objects

    def _create_molecules(self, store: dict) -> None:
        for id_, spec in store.items():
            if isinstance(id_, str):
                id_ = int(id_)

            cls = spec["class"]
            if "parent" in spec:
                spec["parent"] = self[spec["parent"]]

            kwargs = dict(spec)
            del kwargs["class"]
            self._objects[id_] = getattr(aizynthfinder.chem, cls)(**kwargs)


def serialize_action(
    action: RetroReaction, molecule_store: MoleculeSerializer
) -> StrDict:
    """
    Serialize a retrosynthesis action

    :param action: the (re)action to serialize
    :param molecule_store: the molecule serialization object
    :return: the action as a dictionary
    """
    dict_ = action.to_dict()
    dict_["mol"] = molecule_store[dict_["mol"]]
    if not action.unqueried:
        dict_["reactants"] = [
            [molecule_store[item] for item in lst_] for lst_ in action.reactants
        ]
    dict_["class"] = f"{action.__class__.__module__}.{action.__class__.__name__}"
    return dict_


def deserialize_action(
    dict_: StrDict, molecule_store: MoleculeDeserializer
) -> RetroReaction:
    """
    Deserialize a retrosynthesis action

    :param dict_: the (re)action as a dictionary
    :param molecule_store: the molecule deserialization object
    :return: the created action object
    """
    mol_spec = dict_.pop("mol")
    dict_["mol"] = molecule_store.get_tree_molecules([mol_spec])[0]
    try:
        class_spec = dict_.pop("class")
    except KeyError:
        class_spec = "aizynthfinder.chem.TemplatedRetroReaction"
    cls = load_dynamic_class(class_spec)
    if "reactants" in dict_:
        reactants = [
            molecule_store.get_tree_molecules(lst_) for lst_ in dict_.pop("reactants")
        ]
        return cls.from_serialization(dict_, reactants)
    return cls(**dict_)



"""
Helper routines and class for the `aizynthfinder.analysis` package.
To avoid clutter in that package, larger utility algorithms are placed herein.
"""
from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING

import networkx as nx

from aizynthfinder.chem import FixedRetroReaction, Molecule, UniqueMolecule
from aizynthfinder.reactiontree import ReactionTree
from aizynthfinder.utils.image import make_visjs_page

if TYPE_CHECKING:
    from aizynthfinder.utils.type_utils import (
        FrameColors,
        Optional,
        Sequence,
        StrDict,
        Tuple,
    )


@dataclass
class RouteSelectionArguments:
    """
    Selection arguments for the tree analysis class

    If `return_all` is False, it will return at least `nmin` routes and if routes have the same
    score it will return them as well up to `nmax` routes.

    If `return_all` is True, it will return all solved routes if there is at least one is solved, otherwise
    the `nmin` and `nmax` will be used.
    """

    nmin: int = 5
    nmax: int = 25
    return_all: bool = False


class CombinedReactionTrees:
    """
    Encapsulation of an algorithm that combines several reaction trees into a
    larger bipartite graph with all reactions and molecules.

    The reactions at a specific level of the reaction trees are grouped based
    on the reaction smiles.

    :params reactions_trees: the list of reaction trees to combine
    """

    def __init__(self, reaction_trees: Sequence[ReactionTree]) -> None:
        self.graph = nx.DiGraph()
        first_rt = reaction_trees[0]
        self.root = first_rt.root

        self.graph.add_node(self.root, in_stock=first_rt.in_stock(self.root))
        rt_node_spec = [(rt.root, rt.graph) for rt in reaction_trees]
        self._add_reaction_trees_to_node(self.root, rt_node_spec)

    def to_dict(self) -> StrDict:
        """
        Returns the graph as a dictionary in a pre-defined format.

        :return: the combined reaction trees
        """
        rtree = ReactionTree()
        rtree.root = self.root
        rtree.graph = self.graph
        return rtree.to_dict()

    def to_visjs_page(
        self,
        filename: str,
        in_stock_colors: Optional[FrameColors] = None,
    ) -> None:
        """
        Create a visualization of the combined reaction tree using the vis.js network library.

        The HTML page and all the images will be put into a tar-ball.

        :param filename: the name of the tarball
        :param in_stock_colors: the colors around molecules, defaults to {True: "green", False: "orange"}
        """
        in_stock_colors = in_stock_colors or {True: "green", False: "orange"}
        molecules = [node for node in self.graph if isinstance(node, Molecule)]
        reactions = [node for node in self.graph if not isinstance(node, Molecule)]
        frame_colors = [
            in_stock_colors[self.graph.nodes[node].get("in_stock", False)]
            for node in molecules
        ]
        make_visjs_page(filename, molecules, reactions, self.graph.edges, frame_colors)

    def _add_reaction_trees_to_node(
        self,
        base_node: UniqueMolecule,
        rt_node_spec: Sequence[Tuple[UniqueMolecule, nx.DiGraph]],
    ) -> None:
        reaction_groups = defaultdict(list)
        # Group the reactions from the nodes at this level based on the reaction smiles
        for node, graph in rt_node_spec:
            for reaction in graph[node]:
                reaction_groups[reaction.reaction_smiles()].append((graph, reaction))

        for group in reaction_groups.values():
            # Use the first RT in each group as the base
            first_graph, first_reaction = group[0]
            reaction_node = first_reaction.copy()
            self.graph.add_edge(base_node, reaction_node)

            for child in first_graph[first_reaction]:
                mol_node = child.make_unique()
                self.graph.add_node(
                    mol_node, in_stock=first_graph.nodes[child].get("in_stock", False)
                )
                self.graph.add_edge(reaction_node, mol_node)
                self._add_reaction_trees_to_node(
                    mol_node, self._find_other_children(child, group)
                )

    @staticmethod
    def _find_other_children(
        child: UniqueMolecule, group: Sequence[Tuple[nx.DiGraph, FixedRetroReaction]]
    ) -> Sequence[Tuple[UniqueMolecule, nx.DiGraph]]:
        children_spec = []
        for other_graph, other_reaction in group:
            found = False
            for other_child in other_graph[other_reaction]:
                if other_child.inchi_key == child.inchi_key:
                    children_spec.append((other_child, other_graph))
                    found = True
                    break
            if not found:
                raise ValueError("Could not find other child")
        return children_spec



""" Module for base classes for AND/OR trees and some tree utility code """
from __future__ import annotations

import abc
import random
from typing import TYPE_CHECKING

import networkx as nx

from aizynthfinder.chem import UniqueMolecule
from aizynthfinder.reactiontree import ReactionTree, ReactionTreeLoader

if TYPE_CHECKING:
    from aizynthfinder.chem import FixedRetroReaction
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.context.stock import Stock
    from aizynthfinder.utils.type_utils import Any, List, Optional, StrDict, Union


class TreeNodeMixin:
    """A mixin class for node in a tree"""

    @property
    def prop(self) -> StrDict:
        """Dictionary with publicly exposed properties"""
        return {}

    @property
    def children(self) -> List["TreeNodeMixin"]:
        """List of children nodes"""
        return []


class AndOrSearchTreeBase(abc.ABC):
    """A base class for a search tree based on an AND/OR structure"""

    def __init__(
        self, config: Configuration, root_smiles: Optional[str] = None
    ) -> None:
        self.config = config
        self._root_smiles = root_smiles

    @property
    def mol_nodes(self) -> List[TreeNodeMixin]:
        """Return the molecule nodes of the tree"""
        return []

    @abc.abstractmethod
    def one_iteration(self) -> bool:
        """Perform one iteration of the search"""
        return False

    @abc.abstractmethod
    def routes(self) -> List[ReactionTree]:
        """Return the routes of the tree"""
        return []


class SplitAndOrTree:
    """
    Encapsulation of an algorithm to split an AND/OR tree into separate routes

    This is a modified algorithm of the one detailed in the CompRet paper:
    Shibukawa et al. (2020) J. Cheminf. 12, 52

    This implementation sets an upper-bound on the number of extracted routes
    to avoid combinatorial explosion.

    The routes are extracted on instantiation and the routes can be access from
    the `routes` attribute.

    :param root_node: the root of the AND/OR tree
    :param stock: the stock of the search
    :param max_routes: the maximum number of routes to extract
    """

    def __init__(
        self, root_node: TreeNodeMixin, stock: Stock, max_routes: int = 25000
    ) -> None:
        self._traces: List[_AndOrTrace] = []
        self._black_list: List[TreeNodeMixin] = []
        graph = _AndOrTrace(root_node)
        self._samples = {child: 0 for child in root_node.children}
        if root_node.children:
            self._sampling_cutoff = max_routes / len(root_node.children)
        else:
            self._sampling_cutoff = max_routes
        self._partition_search_tree(graph, root_node)
        routes_list = [
            ReactionTreeFromAndOrTrace(trace, stock).tree for trace in self._traces
        ]
        routes_map = {route.hash_key(): route for route in routes_list}
        self.routes = list(routes_map.values())

    def _partition_search_tree(self, graph: _AndOrTrace, node: TreeNodeMixin) -> None:
        # fmt: off
        if self._sampling_cutoff and len(graph) > 1 and self._samples[graph.first_reaction] > self._sampling_cutoff:
            return
        # fmt: on

        if not node.children:
            self._traces.append(graph)

        children_to_search = [
            child for child in node.children if child not in self._black_list
        ]
        if not children_to_search:
            for child in node.children:
                self._black_list.remove(child)
            return

        graph_copy = graph.copy()

        child_node = self._select_child_node(children_to_search)
        graph.add_edge(node, child_node)
        for grandchild in child_node.children:
            graph.add_edge(child_node, grandchild)
        self._black_list.append(child_node)

        leaves = [node for node in graph if not graph[node] and node.children]
        if not leaves:
            self._traces.append(graph)
            self._samples[graph.first_reaction] += 1
        else:
            self._partition_search_tree(graph, leaves[0])

        self._partition_search_tree(graph_copy, node)

    def _select_child_node(self, children: List[TreeNodeMixin]) -> TreeNodeMixin:
        # This is what makes this algorithm different from what was
        # detailed in the CompRet paper
        if not self._sampling_cutoff:
            return children[0]

        solved_children = [
            child for child in children if child.prop.get("solved", False)
        ]
        if solved_children:
            return random.choice(solved_children)
        return random.choice(children)


class _AndOrTrace(nx.DiGraph):
    """Helper class for the SplitAndOrTree class."""

    def __init__(self, root: Optional[TreeNodeMixin] = None) -> None:
        super().__init__()
        self.root = root
        self._first_reaction: Optional[TreeNodeMixin] = None
        self._reaction_tree: Optional[Any] = None
        if root:
            self.add_node(root)

    @property
    def first_reaction(self) -> TreeNodeMixin:
        """Return the first reaction or raise an exception"""
        assert self._first_reaction is not None
        return self._first_reaction

    def add_edge(
        self, u_of_edge: TreeNodeMixin, v_of_edge: TreeNodeMixin, **attr: Any
    ) -> None:
        if u_of_edge is self.root:
            if self._first_reaction is not None:
                raise ValueError(
                    "Re-defining trace. Trying to set first reaction twice."
                )
            self._first_reaction = v_of_edge
        super().add_edge(u_of_edge, v_of_edge, **attr)

    def copy(self, as_view: bool = False) -> "_AndOrTrace":
        other = super().copy(as_view)
        assert isinstance(other, _AndOrTrace)
        other.root = self.root
        other._first_reaction = self._first_reaction  # pylint: disable=protected-access
        return other


class ReactionTreeFromAndOrTrace(ReactionTreeLoader):
    """Creates a reaction tree object from an AND/OR Trace"""

    def _load(self, andor_trace: nx.DiGraph, stock: Stock) -> None:  # type: ignore
        """
        :param andor_trace: the trace graph
        :param stock: stock object
        """
        self._stock = stock
        self._trace_graph = andor_trace
        self._trace_root = self._find_root()

        self._add_node(
            self._unique_mol(self._trace_root.prop["mol"]),
            depth=0,
            transform=0,
            in_stock=self._trace_root.prop["mol"] in self._stock,
        )
        for node1, node2 in andor_trace.edges():
            if "reaction" in node2.prop and not andor_trace[node2]:
                continue
            rt_node1 = self._make_rt_node(node1)
            rt_node2 = self._make_rt_node(node2)
            self.tree.graph.add_edge(rt_node1, rt_node2)

    def _add_node_with_depth(
        self, node: Union[UniqueMolecule, FixedRetroReaction], base_node: TreeNodeMixin
    ) -> None:
        if node in self.tree.graph:
            return

        depth = nx.shortest_path_length(self._trace_graph, self._trace_root, base_node)
        if isinstance(node, UniqueMolecule):
            self._add_node(
                node, depth=depth, transform=depth // 2, in_stock=node in self._stock
            )
        else:
            self._add_node(node, depth=depth)

    def _find_root(self) -> TreeNodeMixin:
        for node, degree in self._trace_graph.in_degree():  # type: ignore
            if degree == 0:
                return node
        raise ValueError("Could not find root!")

    def _make_rt_node(
        self, node: TreeNodeMixin
    ) -> Union[UniqueMolecule, FixedRetroReaction]:
        if "mol" in node.prop:
            unique_obj = self._unique_mol(node.prop["mol"])
            self._add_node_with_depth(unique_obj, node)
            return unique_obj

        reaction_obj = self._unique_reaction(node.prop["reaction"])
        reaction_obj.reactants = (
            tuple(self._unique_mol(child.prop["mol"]) for child in node.children),
        )
        self._add_node_with_depth(reaction_obj, node)
        return reaction_obj



""" Sub-package containing chemistry routines
"""
from aizynthfinder.chem.mol import (
    Molecule,
    MoleculeException,
    TreeMolecule,
    UniqueMolecule,
    none_molecule,
)
from aizynthfinder.chem.reaction import (
    FixedRetroReaction,
    RetroReaction,
    SmilesBasedRetroReaction,
    TemplatedRetroReaction,
    hash_reactions,
)
from aizynthfinder.chem.serialization import (
    MoleculeDeserializer,
    MoleculeSerializer,
    deserialize_action,
    serialize_action,
)



""" Module containing a class for encapsulating the settings of the tree search
"""
from __future__ import annotations

import os
import re
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import yaml

from aizynthfinder.context.policy import ExpansionPolicy, FilterPolicy
from aizynthfinder.context.scoring import ScorerCollection
from aizynthfinder.context.stock import Stock
from aizynthfinder.utils.logging import logger

if TYPE_CHECKING:
    from aizynthfinder.utils.type_utils import Any, Dict, List, Optional, StrDict, Union


@dataclass
class _PostprocessingConfiguration:
    min_routes: int = 5
    max_routes: int = 25
    all_routes: bool = False
    route_distance_model: Optional[str] = None
    route_scorers: List[str] = field(default_factory=lambda: [])
    scorer_weights: Optional[List[float]] = field(default_factory=lambda: None)


@dataclass
class _SearchConfiguration:
    algorithm: str = "mcts"
    algorithm_config: Dict[str, Any] = field(
        default_factory=lambda: {
            "C": 1.4,
            "default_prior": 0.5,
            "use_prior": True,
            "prune_cycles_in_search": True,
            "search_rewards": ["state score"],
            "immediate_instantiation": (),
            "mcts_grouping": None,
            "search_rewards_weights": [],
        }
    )
    max_transforms: int = 6
    iteration_limit: int = 100
    time_limit: int = 120
    return_first: bool = False
    exclude_target_from_stock: bool = True
    break_bonds: List[List[int]] = field(default_factory=list)
    freeze_bonds: List[List[int]] = field(default_factory=list)
    break_bonds_operator: str = "and"


@dataclass
class Configuration:
    """
    Encapsulating the settings of the tree search, including the policy,
    the stock, the loaded scorers and various parameters.
    """

    search: _SearchConfiguration = field(default_factory=_SearchConfiguration)
    post_processing: _PostprocessingConfiguration = field(
        default_factory=_PostprocessingConfiguration
    )
    stock: Stock = field(init=False)
    expansion_policy: ExpansionPolicy = field(init=False)
    filter_policy: FilterPolicy = field(init=False)
    scorers: ScorerCollection = field(init=False)

    def __post_init__(self) -> None:
        self.stock = Stock()
        self.expansion_policy = ExpansionPolicy(self)
        self.filter_policy = FilterPolicy(self)
        self.scorers = ScorerCollection(self)
        self._logger = logger()

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Configuration):
            return False
        for key, setting in vars(self).items():
            if isinstance(setting, (int, float, str, bool, list)):
                if (
                    vars(self)[key] != vars(other)[key]
                    or self.search != other.search
                    or self.post_processing != other.post_processing
                ):
                    return False
        return True

    @classmethod
    def from_dict(cls, source: StrDict) -> "Configuration":
        """
        Loads a configuration from a dictionary structure.
        The parameters not set in the dictionary are taken from the default values.
        The policies and stocks specified are directly loaded.

        :param source: the dictionary source
        :return: a Configuration object with settings from the source
        """
        expansion_config = source.pop("expansion", {})
        filter_config = source.pop("filter", {})
        stock_config = source.pop("stock", {})
        scorer_config = source.pop("scorer", {})

        config_obj = Configuration()
        config_obj._update_from_config(dict(source))

        config_obj.expansion_policy.load_from_config(**expansion_config)
        config_obj.filter_policy.load_from_config(**filter_config)
        config_obj.stock.load_from_config(**stock_config)
        config_obj.scorers.create_default_scorers()
        config_obj.scorers.load_from_config(**scorer_config)

        return config_obj

    @classmethod
    def from_file(cls, filename: str) -> "Configuration":
        """
        Loads a configuration from a yaml file.
        The parameters not set in the yaml file are taken from the default values.
        The policies and stocks specified in the yaml file are directly loaded.
        The parameters in the yaml file may also contain environment variables as
        values.

        :param filename: the path to a yaml file
        :return: a Configuration object with settings from the yaml file
        :raises:
            ValueError: if parameter's value expects an environment variable that
                does not exist in the current environment
        """
        with open(filename, "r") as fileobj:
            txt = fileobj.read()
        environ_var = re.findall(r"\$\{.+?\}", txt)
        for item in environ_var:
            if item[2:-1] not in os.environ:
                raise ValueError(f"'{item[2:-1]}' not in environment variables")
            txt = txt.replace(item, os.environ[item[2:-1]])
        _config = yaml.load(txt, Loader=yaml.SafeLoader)
        return Configuration.from_dict(_config)

    def _update_from_config(self, config: StrDict) -> None:
        self.post_processing = _PostprocessingConfiguration(
            **config.pop("post_processing", {})
        )

        search_config = config.pop("search", {})
        for setting, value in search_config.items():
            if value is None:
                continue
            if not hasattr(self.search, setting):
                raise AttributeError(f"Could not find attribute to set: {setting}")
            if setting.endswith("_bonds"):
                if not isinstance(value, list):
                    raise ValueError("Bond settings need to be lists")
                value = _handle_bond_pair_tuples(value) if value else []
            if setting == "algorithm_config":
                if not isinstance(value, dict):
                    raise ValueError("algorithm_config settings need to be dictionary")
                self.search.algorithm_config.update(value)
            else:
                setattr(self.search, setting, value)


def _handle_bond_pair_tuples(bonds: List[List[int]]) -> List[List[int]]:
    if not all(len(bond_pair) == 2 for bond_pair in bonds):
        raise ValueError("Lists of bond pairs to break/freeze should be of length 2")
    return [bond_pair[:2] for bond_pair in bonds]



""" Module containing helper routines for policies
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from aizynthfinder.chem import TreeMolecule
    from aizynthfinder.chem.reaction import RetroReaction
    from aizynthfinder.utils.type_utils import Any, Union


def _make_fingerprint(
    obj: Union[TreeMolecule, RetroReaction], model: Any, chiral: bool = False
) -> np.ndarray:
    fingerprint = obj.fingerprint(radius=2, nbits=len(model), chiral=chiral)
    return fingerprint.reshape([1, len(model)])



""" Module containing a class that holds the tree search
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from aizynthfinder.reactiontree import ReactionTree
from aizynthfinder.search.andor_trees import AndOrSearchTreeBase, SplitAndOrTree
from aizynthfinder.search.dfpn.nodes import MoleculeNode, ReactionNode
from aizynthfinder.utils.logging import logger

if TYPE_CHECKING:
    from aizynthfinder.chem import RetroReaction
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.search.andor_trees import TreeNodeMixin
    from aizynthfinder.utils.type_utils import List, Optional, Sequence, Union


class SearchTree(AndOrSearchTreeBase):
    """
    Encapsulation of the Depth-First Proof-Number (DFPN) search algorithm.

    This algorithm does not support:
        1. Filter policy
        2. Serialization and deserialization

    :ivar config: settings of the tree search algorithm
    :ivar root: the root node

    :param config: settings of the tree search algorithm
    :param root_smiles: the root will be set to a node representing this molecule, defaults to None
    """

    def __init__(
        self, config: Configuration, root_smiles: Optional[str] = None
    ) -> None:
        super().__init__(config, root_smiles)
        self._mol_nodes: List[MoleculeNode] = []
        self._logger = logger()
        self._root_smiles = root_smiles
        if root_smiles:
            self.root: Optional[MoleculeNode] = MoleculeNode.create_root(
                root_smiles, config, self
            )
            self._mol_nodes.append(self.root)
        else:
            self.root = None

        self._routes: List[ReactionTree] = []
        self._frontier: Optional[Union[MoleculeNode, ReactionNode]] = None
        self._initiated = False

        self.profiling = {
            "expansion_calls": 0,
            "reactants_generations": 0,
        }

    @property
    def mol_nodes(self) -> Sequence[MoleculeNode]:  # type: ignore
        """Return the molecule nodes of the tree"""
        return self._mol_nodes

    def one_iteration(self) -> bool:
        """
        Perform one iteration of expansion.

        If possible expand the frontier node twice, i.e. expanding an OR
        node and then and AND node. If frontier not expandable step up in the
        tree and find a new frontier to expand.

        If a solution is found, mask that tree for exploration and start over.

        :raises StopIteration: if the search should be pre-maturely terminated
        :return: if a solution was found
        :rtype: bool
        """
        if not self._initiated:
            if self.root is None:
                raise ValueError("Root is undefined. Cannot make an iteration")

            self._routes = []
            self._frontier = self.root
        assert self.root is not None

        while True:
            # Expand frontier, should be OR node
            assert isinstance(self._frontier, MoleculeNode)
            expanded_or = self._search_step()
            expanded_and = False
            if self._frontier:
                # Expand frontier again, this time an AND node
                assert isinstance(self._frontier, ReactionNode)
                expanded_and = self._search_step()
            if (
                expanded_or
                or expanded_and
                or self._frontier is None
                or self._frontier is self.root
            ):
                break

        found_solution = any(child.proven for child in self.root.children)
        if self._frontier is self.root:
            self.root.reset()

        if self._frontier is None:
            raise StopIteration()

        return found_solution

    def routes(self) -> List[ReactionTree]:
        """
        Extracts and returns routes from the AND/OR tree

        :return: the routes
        """
        if self.root is None:
            return []
        if not self._routes:
            self._routes = SplitAndOrTree(self.root, self.config.stock).routes
        return self._routes

    def _search_step(self) -> bool:
        assert self._frontier is not None
        expanded = False
        if self._frontier.expandable:
            self._frontier.expand()
            expanded = True
            if isinstance(self._frontier, ReactionNode):
                self._mol_nodes.extend(self._frontier.children)

        self._frontier.update()
        if not self._frontier.explorable():
            self._frontier = self._frontier.parent
            return False

        child = self._frontier.promising_child()
        if not child:
            self._frontier = self._frontier.parent
            return False

        self._frontier = child
        return expanded
